//
// hello.ss
// Hello World in SurgeScript
// Copyright 2017 Alexandre Martins <alemartf(at)gmail(dot)com>
//

object "Application"
{
    state "main"
    {
        Console.print("Hello, world!");
        Application.exit();
    }
}